import dataclasses
from collections.abc import Mapping, Sequence
from types import SimpleNamespace

import pytest
import torch

import delu

from .util import Point, PointDC


def flatten(data):
    if isinstance(data, torch.Tensor):
        yield data
    elif isinstance(data, (str, bytes)):
        # mypy: NaN
        yield data  # type: ignore
    elif isinstance(data, Sequence):
        for x in data:
            yield from flatten(x)
    elif isinstance(data, Mapping):
        for x in data.values():
            yield from flatten(x)
    elif isinstance(data, SimpleNamespace):
        for x in vars(data).values():
            yield from flatten(x)
    elif dataclasses.is_dataclass(data):
        for x in vars(data).values():
            yield from flatten(x)
    else:
        yield data


def test_to():
    with pytest.raises(ValueError):
        delu.to(None)
    with pytest.raises(ValueError):
        delu.to([None, None])

    t = lambda x: torch.tensor(0, dtype=x)  # noqa
    f32 = torch.float32
    i64 = torch.int64

    for dtype in f32, i64:
        x = t(dtype)
        assert delu.to(x, dtype) is x
    assert delu.to(t(f32), i64).dtype is i64

    for Container in tuple, Point, list:
        constructor = Container._make if Container is Point else Container
        for dtype in [f32, i64]:
            x = constructor([t(f32), t(f32)])
            out = delu.to(x, dtype)
            assert isinstance(out, Container)
            assert all(x.dtype is dtype for x in out)
            if dtype is f32:
                for x, y in zip(out, x):
                    assert x is y

    data = [t(f32), t(f32)]
    for x, y in zip(delu.to(data, f32), data):
        assert x is y
    assert all(x.dtype is i64 for x in delu.to(data, i64))

    @dataclasses.dataclass
    class A:
        a: torch.Tensor

    data = {
        'a': [t(f32), (t(f32), t(f32))],
        'b': {'c': {'d': [[[t(f32)]]]}},
        'c': Point(t(f32), {'d': t(f32)}),
        'f': A(a=t(f32)),
    }
    for x, y in zip(flatten(delu.to(data, f32)), flatten(data)):
        assert x is y
    for x, y in zip(flatten(delu.to(data, i64)), flatten(data)):
        assert x.dtype is i64
        assert type(x) is type(y)


def test_cat():
    with pytest.raises(ValueError):
        delu.cat([])

    with pytest.raises(ValueError):
        delu.cat([0])

    with pytest.raises(ValueError):
        delu.cat([[0], [1]])

    # simple cases are alredy tested in the doctests

    sequence = [
        Point(torch.tensor([[0, 1]]), torch.tensor([[[0.0, 1.0]]])),
        Point(torch.tensor([[2, 3]]), torch.tensor([[[2.0, 3.0]]])),
    ]

    actual0 = delu.cat(sequence)
    assert isinstance(actual0, Point)
    assert torch.equal(actual0.x, torch.tensor([[0, 1], [2, 3]]))
    assert torch.equal(actual0.y, torch.tensor([[[0.0, 1.0]], [[2.0, 3.0]]]))

    actual1 = delu.cat(sequence, dim=1)
    assert isinstance(actual1, Point)
    assert torch.equal(actual1.x, torch.tensor([[0, 1, 2, 3]]))
    assert torch.equal(actual1.y, torch.tensor([[[0.0, 1.0], [2.0, 3.0]]]))


@pytest.mark.parametrize('batch_size', list(range(1, 11)))
def test_iter_batches(batch_size):
    def check(batches, correct):
        sizes = list(map(len, batches))
        assert sum(sizes) == len(correct)
        assert (
            set(sizes) == {batch_size}
            or set(sizes[:-1]) == {batch_size}
            and sizes[-1] == len(correct) % batch_size
        )
        assert torch.equal(torch.cat(batches), correct)

    n = 10

    # test batch size only
    data = torch.arange(n)
    batches = list(delu.iter_batches(data, batch_size))
    check(batches, data)

    data = Point(torch.arange(n), torch.arange(n, n * 2))
    batches = list(delu.iter_batches(data, batch_size))
    assert all(isinstance(t, Point) for t in batches)
    for i in range(2):
        batches_i = list(x[i] for x in batches)
        check(batches_i, data[i])

    data = (torch.arange(n), torch.arange(n, n * 2))
    batches = list(delu.iter_batches(data, batch_size))
    for i in range(2):
        batches_i = list(x[i] for x in batches)
        check(batches_i, data[i])

    data = {'a': torch.arange(n), 'b': torch.arange(n, n * 2)}
    batches = list(delu.iter_batches(data, batch_size))
    for key in data:
        batches_i = list(x[key] for x in batches)
        check(batches_i, data[key])

    assert (
        len(list(delu.iter_batches(torch.arange(n), batch_size, drop_last=True)))
        == n // batch_size
    )


def test_iter_batches_shuffle():
    a = torch.arange(1000)

    with pytest.raises(ValueError):
        next(delu.iter_batches(a, 10, generator=torch.Generator()))

    batches = list(delu.iter_batches(a, 10, shuffle=True))
    assert not torch.equal(torch.cat(batches), a)
    assert sorted(torch.cat(batches).tolist()) == a.tolist()

    gen = torch.Generator()
    state = gen.get_state()
    batches0 = list(delu.iter_batches(a, 10, shuffle=True, generator=gen))
    gen.set_state(state)
    batches1 = list(delu.iter_batches(a, 10, shuffle=True, generator=gen))
    for x in zip(batches0, batches1):
        assert torch.equal(*x)


def test_iter_batches_bad_input():
    # empty input
    with pytest.raises(ValueError):
        next(delu.iter_batches(torch.empty(0), 1))
    with pytest.raises(ValueError):
        next(delu.iter_batches((), 1))
    with pytest.raises(ValueError):
        next(delu.iter_batches({}, 1))

    @dataclasses.dataclass
    class BadPoint:
        pass

    with pytest.raises(ValueError):
        next(delu.iter_batches(BadPoint(), 1))

    # different lengths
    a = torch.tensor([0, 1])
    b = torch.tensor([0, 1, 2])
    with pytest.raises(ValueError):
        next(delu.iter_batches((a, b), 2))
    with pytest.raises(ValueError):
        next(delu.iter_batches({'a': a, 'b': b}, 2))
    with pytest.raises(ValueError):
        next(delu.iter_batches(PointDC(a, b), 2))



import random

import numpy as np
import torch
from pytest import raises

import delu

from .util import requires_gpu


def _test_seed(functions):
    high = 1000000
    for seed in range(10):
        x = [None, None]
        for i in range(2):
            delu.random.seed(seed)
            x[i] = [f(high) for f in functions]
        assert x[0] == x[1]


def test_seed_cpu():
    _test_seed(
        [
            lambda x: random.randint(0, x),
            lambda x: np.random.randint(x),
            lambda x: torch.randint(x, (1,))[0].item(),
        ]
    )
    for seed in [-1, 2**33]:
        with raises(AssertionError):
            delu.random.seed(seed)


@requires_gpu
def test_seed_gpu():
    functions = []
    for i in range(torch.cuda.device_count()):

        def f(x):
            return (torch.randint(x, (1,), device=f'cuda:{i}')[0].item(),)

        functions.append(f)
    _test_seed(functions)


def test_get_set_state():
    high = 1000000

    def f():
        size = (1,)
        x = (
            random.randint(0, high),
            np.random.randint(0, high, size),
            torch.randint(0, high, size),
        )
        if torch.cuda.is_available():
            x += tuple(
                torch.randint(0, high, size, device=f'cuda:{d}')
                for d in range(torch.cuda.device_count())
            )
        return x

    state = delu.random.get_state()
    value = f()
    for _ in range(10):
        delu.random.set_state(state)
        assert value == f()

    if torch.cuda.is_available():
        state['torch.cuda'] = []
        with raises(AssertionError):
            delu.random.set_state(state)
    else:
        state['torch.cuda'] = [None]
        with raises(AssertionError):
            delu.random.set_state(state)



from collections import namedtuple
from dataclasses import dataclass

import pytest
import torch

Point = namedtuple('Point', ['x', 'y'])
Model = namedtuple('Model', ['model', 'weight', 'bias', 'loss_fn', 'optimizer', 'data'])


@dataclass
class PointDC:
    x: torch.Tensor
    y: torch.Tensor


ignore_deprecated_warning = pytest.mark.filterwarnings('ignore:.*deprecated.*')


class ObjectCounter:
    def __init__(self, sign):
        self.sign = sign
        self.reset()

    def reset(self):
        self.count = 0
        return self

    def update(self, data):
        self.count += len(data[0])
        return self

    def compute(self):
        assert not self.empty
        return self.sign * self.count

    @property
    def empty(self):
        return not self.count


def make_model(data):
    model = torch.nn.Linear(data.shape[1], 1)
    return Model(
        model,
        model.weight.clone(),
        model.bias.clone(),
        lambda: model(data).sum(),
        torch.optim.SGD(model.parameters(), 0.0001),
        data,
    )


requires_gpu = pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required for this test"
)



import datetime
import platform
import sys
from pathlib import Path

# add the repository root to PYTHONPATH
delu_path = Path.cwd()
while not (delu_path.name == 'delu' and delu_path.parent.name != 'delu'):
    delu_path = delu_path.parent
sys.path.append(str(delu_path))
import delu  # noqa

# >>> project >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
author = 'Yura52'
copyright = f'{datetime.datetime.now().year}, {author}'
project = 'DeLU'
release = delu.__version__
version = delu.__version__

# >>> general >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
default_role = 'py:obj'
pygments_style = 'default'
repo_url = 'https://github.com/Yura52/delu'
templates_path = ['_templates']

# >>> extensions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_copybutton',
]
if platform.machine() != 'arm64':
    # libenchant is not available for Apple CPUs
    extensions.append('sphinxcontrib.spelling')

autodoc_typehints = 'description'
# autodoc_typehints_format = 'fully-qualified'

doctest_global_setup = '''
import numpy as np
import torch
import delu
'''

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'torch': ('https://pytorch.org/docs/stable', None),
}

napoleon_numpy_docstring = False
napoleon_use_admonition_for_examples = False

spelling_show_suggestions = True

# >>> HTML >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
html_static_path = ['_static']
html_favicon = 'images/favicon.ico'

# >>> Furo >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
extensions.append('sphinx_remove_toctrees')
remove_from_toctrees = ['reference/api/*']

html_title = project
html_css_files = ['furo.css']
html_logo = 'images/logo_120x120.svg'
html_theme = 'furo'
html_theme_options = {
    'sidebar_hide_name': True,
    'source_repository': 'https://github.com/Yura52/delu/',
    'source_branch': 'main',
    'source_directory': 'docs/',
    'footer_icons': [
        {
            'name': 'GitHub',
            'url': repo_url,
            'html': '''
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            ''',  # noqa: E501
            'class': '',
        },
    ],
    'light_css_variables': {
        'admonition-title-font-size': '90.5%',
        'admonition-font-size': '90.5%',
        'code-font-size': 'var(--font-size--small)',
    },
}


# >>> Material Sphinx >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# This theme concatenates autosummary tables nicely and supports versioning.
# import sphinx_material

# html_theme = 'sphinx_material'
# html_css_files = ['material_theme.css']
# html_logo = 'images/logo_28x28.svg'
# html_theme_options = {
#     # Full list of options: https://github.com/bashtage/sphinx-material/blob/master/sphinx_material/sphinx_material/theme.conf
#     'base_url': 'https://yura52.github.io/delu',
#     # Full list of colors (not all of them are available in sphinx-material, see theme.conf above):  # noqa: E501
#     # https://squidfunk.github.io/mkdocs-material/setup/changing-the-colors/#primary-color
#     # Nice colors: white, blue, red, deep purple (indigo in mkdocs)
#     'color_primary': 'white',
#     'globaltoc_depth': 1,
#     # search here for logo icons: https://www.compart.com/en/unicode
#     # 'logo_icon': '&#127968;',
#     'nav_links': [],
#     'nav_title': project + ' ' + version,
#     'repo_name': project,
#     'repo_url': repo_url,
#     'repo_type': 'github',
#     'master_doc': False,
#     'version_dropdown': True,
#     'version_json': '_static/versions.json',
# }
# html_sidebars = {
#     '**': ['logo-text.html', 'globaltoc.html', 'searchbox.html'],
# }

# >>> Read the Docs >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# extensions.append('sphinx_rtd_theme')
# html_theme = 'sphinx_rtd_theme'
# html_css_files = ['rtd_theme.css']
# html_logo = 'images/logo_120x120.svg'

# github_url = repo_url
# html_theme_options = {
#     'logo_only': True,
#     'display_version': False,
#     'prev_next_buttons_location': None,
#     'style_external_links': True,
# }



import math
from copy import deepcopy
from typing import Any, Dict, Iterable, Iterator, Optional, Sized, Union

from ._utils import deprecated

_DEFAULT_PROGRESS_BAR_CONFIG: Dict[str, Any] = {}


def _try_len(x):
    try:
        return len(x)
    except (TypeError, NotImplementedError):
        return None


@deprecated('Currently, no alternatives are available.')
class Stream:
    """Smart wrapper for data loaders and iterables.

    `Stream` simplifies managing loops, especially in typical Deep Learning scenarios.
    `Stream`:

    - manages the "epoch" and "iteration" variables
    - allows to dump and restore loop's state: epoch, iteration, etc.
    - allows to customize the size of epoch
    - allows to change the underlying data loader on the fly
    - enables useful patterns

    .. rubric:: Tutorial

    Let's start with the most common training loop::

        loader = DataLoader(...)
        iteration = 0
        for epoch in range(max_epoch):
            for batch in loader:
                iteration += 1
                print('Epoch:', epoch, 'Iteration:', iteration)
                ...

    Let's enhance the loop using `Stream`::

        stream = delu.Stream(DataLoader(...))  # (A)
        for epoch in stream.epochs(max_epoch):  # (B)
            for batch in epoch:  # (C)
                print('Epoch:', stream.epoch, 'Iteration:', stream.iteration)  # (D)
                ...

    Some comments for the above code:

    - :code:`(A)` `Stream` is created by passing a dataloader as a single argument (in fact, you can pass any iterable object);
      the dataloader is accessible via `Stream.loader`
    - :code:`(B)` :code:`epoch` is an iterator over batches for one epoch
    - :code:`(C)` a progress bar for batches is displayed (for the whole training loop, not just for one epoch)
    - :code:`(D)` `Stream.epoch` and `Stream.iteration` are managed automatically

    Saving the loop's state and resuming the loop is possible with the methods
    `Stream.state_dict`, `Stream.load_state_dict`. In practice, it can look like this::

        model = ...
        optimizer = ...
        stream = delu.Stream(DataLoader(...))
        if load_from_checkpoint:
            checkpoint = torch.load(checkpoint_path)
            model.load_state_dict(checkpoint['model'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            stream.load_state_dict(checkpoint['stream'])
        ...
        for epoch in stream.epochs(...):
            for batch in epoch:
                ...
            torch.save(
                {
                    'model': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                    'stream': stream.state_dict(),
                },
                f'checkpoint_{stream.epoch}.pt'
            )

    Note:
        Stream's state does not include the loader's state. See `Stream.state_dict` and
        `Stream.load_state_dict` for details.

    In order to customize the epoch size, pass the size as the second argument::

        for epoch in stream.epochs(max_epoch, custom_epoch_size):
            for batch in epoch:
                ...

    Changing the underlying loader on the fly is possible at *any* moment (even in the
    middle of epoch) via `Stream.set_loader`. For example::

        for epoch in stream.epochs(max_epoch, custom_epoch_size):
            for batch in epoch:
                ...
                if need_new_data():
                    stream.set_loader(new_loader)

    If the method `Stream.epochs` does not fit your workflow and you want more control
    over the loop, there are more "low-level" methods (in fact, `Stream.epochs` is just
    a thin wrapper around them):

    - `Stream.increment_epoch`
    - `Stream.data`
    - `Stream.next`

    For example, the most common training loop can be implemented as follows::

        # A
        while stream.epoch < max_epoch:
            stream.increment_epoch()
            for batch in stream.data():
                ...

        # B
        while stream.epoch < max_epoch:
            stream.increment_epoch()
            for _ in range(len(stream.loader)):
                batch = stream.next()  # stream.iteration is incremented automatically
                ...

    The "infinite" stream of data can be implemented as follows::

        for item in stream.data(float('inf')):
            ...
            if condition:  # for example: `if stream.iteration % frequency == 0`
                ...

    Note:
        For better technical understanding, keep in mind that `Stream` simply
        encapsulates an "infinite iterator" that is constantly moving forward. The
        behavior is absolutely the same for both finite and infinite iterables and can
        be expressed with the following loop::

            while True:
                for item in loader:  # the loader which is passed to the constructor
                    ...

        Documentation for `Stream.next` and `Stream.data` provide helpful examples.
    """  # noqa: E501

    class _EpochData:
        def __init__(self, stream, size):
            self._stream = stream
            self._size = size
            self._start = self._stream.iteration

        def __iter__(self):
            return self

        def __next__(self):
            if (
                self._size is not None
                and self._stream.iteration - self._start >= self._size
            ):
                raise StopIteration()
            return self._stream.next()

    def __init__(self, loader: Iterable) -> None:
        """Initialize self.

        Args:
            loader: any kind of iterable (DataLoader, list, iterator, generator, ...)
        Raises:
            AssertionError: if :code:`loader` is not an iterator and is empty

        Examples:
            .. testcode::

                stream = delu.Stream([0, 1, 2, 3])
                stream = delu.Stream(range(10))
                import itertools
                stream = delu.Stream(itertools.repeat(0))

                from torch.utils.data import DataLoader, TensorDataset
                dataset = TensorDataset(torch.randn(10, 2))
                stream = delu.Stream(DataLoader(dataset, batch_size=3, shuffle=True))
        """
        assert _try_len(loader) != 0
        self._iteration = 0
        self._epoch = 0
        self._loader = loader
        self._iter: Optional[Iterator] = None
        self._progress_bar = None
        self._should_update_progress_bar = False

    @property
    def iteration(self) -> int:
        """Current iteration.

        Technically, the number of `Stream.next` calls.
        """
        return self._iteration

    @property
    def epoch(self) -> int:
        """Current epoch.

        Technically, the number of `Stream.increment_epoch` calls.
        """
        return self._epoch

    @property
    def loader(self) -> Iterable:
        """The underlying loader."""
        return self._loader

    def set_loader(self, loader: Iterable) -> None:
        """Set new loader.

        Args:
            loader:
        Raises:
            AssertionError: if :code:`loader` is not an iterator and is empty.

        Examples:
            .. testcode::

                from itertools import repeat
                stream = delu.Stream(repeat(0))
                for x in stream.data(5):
                    print(stream.iteration, x)
                    if stream.iteration == 2:
                        stream.set_loader(repeat(1))

            .. testoutput::

                1 0
                2 0
                3 1
                4 1
                5 1
        """
        assert _try_len(loader) != 0
        self._loader = loader
        if self._iter is not None:
            self._iter = iter(loader)

    def _increment_iteration(self):
        self._iteration += 1

    def increment_epoch(self) -> None:
        """Increment `Stream.epoch`.

        Examples:
            .. testcode::

                stream = delu.Stream(range(5))
                assert stream.epoch == 0
                stream.increment_epoch()
                assert stream.epoch == 1
                stream.increment_epoch()
                assert stream.epoch == 2
        """
        self._epoch += 1

    def reload_iterator(self) -> None:
        """Set the underlying iterator to `iter(self.loader)`.

        If the underlying loader is a finite iterable, the method can be used to
        interrupt and skip the current epoch (i.e. skip its data). If the loader is an
        iterator, the method does nothing.

        Examples:
            .. testcode::

                stream = delu.Stream(range(5))
                assert stream.next() == 0
                assert stream.next() == 1
                stream.reload_iterator()
                assert stream.next() == 0

                stream = delu.Stream(iter(range(5)))
                assert stream.next() == 0
                assert stream.next() == 1
                stream.reload_iterator()
                assert stream.next() == 2
        """
        self._iter = iter(self.loader)

    def next(self) -> Any:
        """Get the next item and increment iteration.

        Returns:
            The next item.
        Raises:
            StopIteration: if :code:`loader` is a finite iterator and the data is over

        Examples:
            .. testcode::

                stream = delu.Stream(range(3))
                assert stream.iteration == 0
                assert stream.next() == 0
                assert stream.iteration == 1
                assert stream.next() == 1
                assert stream.next() == 2
                assert stream.next() == 0
                assert stream.iteration == 4

            .. code-block::

                while True:
                    x = stream.next()
                    ...
                    if stream.iteration % frequency:
                        ...
        """
        if self._iter is None:
            self._iter = iter(self._loader)
        if self._progress_bar is not None:
            if self._should_update_progress_bar:
                self._progress_bar.update()
            self._should_update_progress_bar = True
        try:
            value = next(self._iter)
        except StopIteration:
            self.reload_iterator()
            # If the following line raises StopIteration too, then the data is over
            # and the exception should be just propagated.
            value = next(self._iter)
        self._increment_iteration()
        return value

    def data(self, n_items: Optional[Union[int, float]] = None) -> Iterator:
        """Iterate over the loader.

        Under the hood, `Stream.next` is called, hence, `Stream.iteration` changes
        during iterations.

        Args:
            n_items: how many items to produce. If `None`, interpreted as
                :code:`len(self.loader)`. If `float`, must be :code:`float('inf')` or
                `math.inf`.
        Raises:
            AssertionError: if :code:`n_items` is a finite float or nan.
            ValueError: if :code:`loader` is an iterator and :code:`n_items` is
                `None`

        Examples:
            .. testcode::

                stream = delu.Stream(range(5))
                assert list(stream.data()) == [0, 1, 2, 3, 4]
                assert list(stream.data(3)) == [0, 1, 2]
                # stream doesn't "start over"!
                assert list(stream.data(3)) == [3, 4, 0]
                assert list(stream.data(1)) == [1]
                assert list(stream.data(2)) == [2, 3]

            .. code-block::

                for x in stream.data(float('inf')):
                    ...
                    if stream.iteration % frequency:
                        ...
        """
        if isinstance(n_items, float):
            assert math.isinf(n_items)
        if n_items is None:
            if not isinstance(self.loader, Sized):
                raise ValueError()
            n_items = len(self.loader)
        return Stream._EpochData(self, n_items)

    def epochs(
        self,
        max_epoch: Union[int, float],
        epoch_size: Optional[Union[int, float]] = None,
        progress_bar_config: Optional[Dict[str, Any]] = _DEFAULT_PROGRESS_BAR_CONFIG,
    ) -> Iterator[Iterator[Any]]:
        """Iterate over data epochs.

        A shortcut for what is probably the most popular form of a training loop in Deep
        Learning (plus a progress bar)::

            for epoch in stream.epochs(max_epoch, epoch_size):
                for batch in epoch:
                    ...

            # is equivalent to:

            while stream.epoch < max_epoch:
                stream.increment_epoch()
                for batch in stream.data(epoch_size):
                    ...

        Args:
            max_epoch: defines the number of epochs. The loop keeps running while
                :code:`self.epoch < max_epoch`. If `float`, must be :code:`float('inf')`
                or `math.inf`.
            epoch_size: the number of data items in one epoch
                (is forwarded to `Stream.data`).
            progress_bar_config: if not `None` (the default value is :code:`{}`!), a
                progress bar for iterations will be displayed and the argument will be
                interpreted as key-word arguments for
                `tqdm <https://tqdm.github.io/docs/tqdm/#__init__>`_. The following
                key-word arguments will be automatically added if not presented in
                :code:`progress_bar_config`: :code:`initial`, :code:`total` (if can be
                inferred from the arguments and/or from `Stream.loader`). If
                [`ipywidgets`](https://ipywidgets.readthedocs.io) is installed and the
                program is running in JupyterLab (Jupyter Notebook), then the pretty
                widget is used instead of the text-based one.

        Returns:
            Iterator over iterators over data from `Stream.loader`.
        Raises:
            AssertionError: if :code:`max_epoch` is a finite float or nan.

        Examples:
            .. testcode::

                stream = delu.Stream(range(3))
                for epoch in stream.epochs(2):
                    for x in epoch:
                        print(x)
                    print('-')

            .. testoutput::

                0
                1
                2
                -
                0
                1
                2
                -

            .. testcode::

                stream = delu.Stream(range(3))
                for epoch in stream.epochs(3, 2):
                    for x in epoch:
                        print(x)
                    print('-')

            .. testoutput::

                0
                1
                -
                2
                0
                -
                1
                2
                -
        """
        if isinstance(max_epoch, float):
            assert math.isinf(max_epoch)
        progress_bar_config = None
        if progress_bar_config is not None:
            raise RuntimeError('unreachable')
            if progress_bar_config is _DEFAULT_PROGRESS_BAR_CONFIG:
                # the default value is MUTABLE, so let's deepcopy it
                # just in case it is modified later
                progress_bar_config = deepcopy(progress_bar_config)
            pbar_epoch_size = (
                _try_len(self.loader) if epoch_size is None else epoch_size
            )
            all_progress_bar_options = {
                'initial': self.iteration,
                'total': (
                    None
                    if (pbar_epoch_size is None or math.isinf(max_epoch))
                    else max_epoch * pbar_epoch_size
                ),
            }
            all_progress_bar_options.update(progress_bar_config)
            self._progress_bar = tqdm(**all_progress_bar_options)  # type: ignore  # noqa: F821,E501
            self._should_update_progress_bar = False
        while self.epoch < max_epoch:
            self.increment_epoch()
            yield self.data(epoch_size)
        if self._should_update_progress_bar:
            assert self._progress_bar is not None
            self._progress_bar.update()

    def state_dict(self) -> Dict[str, Any]:
        """Get the stream's state.

        The result can be passed to `Stream.load_state_dict`. The result includes:

        - epoch
        - iteration

        Note:
            Fields related to data (loader, iterator etc.) are **NOT** included in the
            state. If you want to save the "state of data stream" then you have to save
            the state of corresponding random number generators separately.

        Returns:
            state

        See also:
            `Stream.load_state_dict`

        Examples:
            .. testcode::

                stream = delu.Stream(range(10))
                assert stream.state_dict() == {'epoch': 0, 'iteration': 0}
                stream.next()
                stream.next()
                stream.increment_epoch()
                assert stream.state_dict() == {'epoch': 1, 'iteration': 2}
        """
        return {'iteration': self.iteration, 'epoch': self.epoch}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Load state dictionary.

        Args:
            state_dict: state. Must be produced by `Stream.state_dict`.

        Note:
            The method does not affect data that is produced by `Stream.epochs`,
            `Stream.data`, `Stream.next` (see the examples below), i.e. the method
            only sets some "metadata" such as epoch, iteration etc. If you want to
            load the "state of data stream", you have to load the state of corresponding
            random number generators separately.

        See also:
            `Stream.state_dict`

        Examples:

            .. testcode::

                stream = delu.Stream(range(10))
                stream.next()
                stream.increment_epoch()
                assert stream.state_dict() == {'epoch': 1, 'iteration': 1}

                new_stream = delu.Stream(range(10))
                new_stream.load_state_dict(stream.state_dict())
                assert new_stream.state_dict() == {'epoch': 1, 'iteration': 1}
                assert new_stream.next() == 0
                assert new_stream.state_dict() == {'epoch': 1, 'iteration': 2}
        """
        self._iteration = state_dict['iteration']
        self._epoch = state_dict['epoch']



from pytest import raises

import delu


def test_lambda():
    assert delu.nn.Lambda(lambda: 0)() == 0
    assert delu.nn.Lambda(lambda x: x)(1) == 1
    assert delu.nn.Lambda(lambda x, y, z: x + y + z)(1, 2, z=3) == 6
    with raises(TypeError):
        delu.nn.Lambda(lambda x: x)()



import functools
import inspect
import warnings

from .exceptions import DeLUDeprecationWarning


def deprecated(message: str):
    def decorator(item):
        assert item.__doc__ is not None
        docstring = f'[**DEPRECATED** | {message}]' + ' ' + item.__doc__

        def warn(item_type):
            warnings.warn(
                f'The {item_type} {item.__qualname__}` is deprecated'
                ' and will be removed in future releases. ' + message,
                DeLUDeprecationWarning,
            )

        if isinstance(item, type):
            wrapper = item
            wrapper.__doc__ = docstring
        else:

            assert inspect.isfunction(item)

            @functools.wraps(item)
            def wrapper(*args, **kwargs):
                warn('function')
                return item(*args, **kwargs)

            wrapper.__doc__ = docstring

        return wrapper

    return decorator


def is_namedtuple(x) -> bool:
    return isinstance(x, tuple) and all(
        hasattr(x, attr) for attr in ['_make', '_asdict', '_replace', '_fields']
    )



import gc
from unittest.mock import patch

import pytest
import torch

import delu


@pytest.mark.parametrize('gpu', [False, True])
def test_free_memory(gpu):
    with (patch('gc.collect')) as _, (patch('torch.cuda.empty_cache')) as _, (
        patch('torch.cuda.synchronize')
    ) as _, (patch('torch.cuda.is_available', lambda: gpu)) as _:
        delu.cuda.free_memory()
        gc.collect.assert_called_once()
        if gpu:
            torch.cuda.synchronize.assert_called_once()
            torch.cuda.empty_cache.assert_called_once()



import dataclasses
import math
from typing import Iterable, Iterator, Optional, TypeVar

import torch

from ._utils import deprecated, is_namedtuple

T = TypeVar('T')
K = TypeVar('K')


def to(data: T, *args, **kwargs) -> T:
    """Like `torch.Tensor.to`, but for collections of tensors.

    The function takes a (nested) collection of tensors
    and creates its copy where all the tensors are transformed
    with `torch.Tensor.to`.

    Args:
        data: the tensor or the (nested) collection of tensors. Allowed collections
            include: (named)tuples, lists, dictionaries and dataclasses.
            For dataclasses, all their fields must be tensors.
        args: the positional arguments for `torch.Tensor.to`
        kwargs: the key-word arguments for `torch.Tensor.to`
    Returns:
        transformed data.

    Examples:
        .. testcode::

            # in practice, this can be 'cuda' or any other device
            device = torch.device('cpu')
            tensor = torch.tensor

            x = tensor(0)
            x = delu.to(x, dtype=torch.float, device=device)

            batch = {
                'x': tensor([0.0, 1.0]),
                'y': tensor([0, 1]),
            }
            batch = delu.to(batch, device)

            x = [
                tensor(0.0),
                {'a': tensor(1.0), 'b': tensor(2.0)},
                (tensor(3.0), tensor(4.0))
            ]
            x = delu.to(x, torch.half)
    """

    def TO_(x):
        return to(x, *args, **kwargs)

    # mypy does not understand what is going on here, hence a lot of "type: ignore"
    if isinstance(data, torch.Tensor):
        return data.to(*args, **kwargs)  # type: ignore
    elif isinstance(data, (tuple, list)):
        constructor = type(data)._make if is_namedtuple(data) else type(data)  # type: ignore  # noqa: E501
        return constructor(TO_(x) for x in data)  # type: ignore
    elif isinstance(data, dict):
        return type(data)((k, TO_(v)) for k, v in data.items())  # type: ignore
    elif dataclasses.is_dataclass(data):
        return type(data)(**{k: TO_(v) for k, v in vars(data).items()})  # type: ignore
    else:
        raise ValueError(
            f'the input contains an object of the unsupported type {type(data)}.'
            ' See the documentation for details'
        )


def cat(iterable: Iterable[T], dim: int = 0) -> T:
    """Like `torch.cat`, but for collections of tensors.

    The function is especially useful for concatenating outputs of a function or a model
    that returns not a single tensor, but a (named)tuple/dictionary/dataclass
    of tensors. For example::

        class Model(nn.Module):
            ...
            def forward(...) -> tuple[Tensor, Tensor]:
                ...
                return y_pred, embeddings

        model = Model(...)
        dataset = Dataset(...)
        dataloader = DataLoader(...)

        # prediction
        model.eval()
        with torch.inference_mode():
            # Concatenate a sequence of tuples into a single tuple.
            y_pred, embeddings = delu.cat(model(batch) for batch in dataloader)
        assert isinstance(y_pred, torch.Tensor) and len(y_pred) == len(dataset)
        assert isinstance(embeddings, torch.Tensor) and len(embeddings) == len(dataset)

    See other examples below.

    Note:
        Under the hood, roughly speaking, the function "transposes" the sequence of
        collections to a collection of sequences and applies `torch.cat` to those
        sequencies.

    Args:
        iterable: the iterable of (named)tuples/dictionaries/dataclasses of tensors.
            All items of the iterable must be of the same type and have the same
            structure (tuples must be of the same length, dictionaries must have the
            same keys, dataclasses must have the same fields). Dataclasses must have
            only tensor-valued fields.
        dim: the argument for `torch.cat`.
    Returns:
        Concatenated items of the iterable.
    Raises:
        ValueError: if the iterable is empty or contains unsupported collections.

    See also:
        `delu.iter_batches`

    Examples:
        Below, only one-dimensional data and dim=0 are considered for simplicity.

        .. testcode::

            tensor = torch.tensor

            batches = [
                # (batch0_x,         batch0_y)
                (tensor([0.0, 1.0]), tensor([[0], [1]])),

                # (batch1_x,         batch1_y)
                (tensor([2.0, 3.0]), tensor([[2], [3]])),

                # ...
            ]
            # result = (x, y)
            result = delu.cat(batches)
            assert isinstance(result, tuple) and len(result) == 2
            assert torch.equal(result[0], tensor([0.0, 1.0, 2.0, 3.0]))
            assert torch.equal(result[1], tensor([[0], [1], [2], [3]]))

            batches = [
                {'x': tensor([0.0, 1.0]), 'y': tensor([[0], [1]])},
                {'x': tensor([2.0, 3.0]), 'y': tensor([[2], [3]])},
            ]
            result = delu.cat(batches)
            assert isinstance(result, dict) and set(result) == {'x', 'y'}
            assert torch.equal(result['x'], tensor([0.0, 1.0, 2.0, 3.0]))
            assert torch.equal(result['y'], tensor([[0], [1], [2], [3]]))

            from dataclasses import dataclass
            @dataclass
            class Data:
                # all fields must be tensors
                x: torch.Tensor
                y: torch.Tensor

            batches = [
                Data(tensor([0.0, 1.0]), tensor([[0], [1]])),
                Data(tensor([2.0, 3.0]), tensor([[2], [3]])),
            ]
            result = delu.cat(batches)
            assert isinstance(result, Data)
            assert torch.equal(result.x, tensor([0.0, 1.0, 2.0, 3.0]))
            assert torch.equal(result.y, tensor([[0], [1], [2], [3]]))

            x = tensor([0.0, 1.0, 2.0, 3.0, 4.0])
            y = tensor([[0], [10], [20], [30], [40]])
            batch_size = 2
            ab = delu.cat(delu.iter_batches((x, y), batch_size))
            assert torch.equal(ab[0], x)
            assert torch.equal(ab[1], y)
    """
    data = iterable if isinstance(iterable, list) else list(iterable)
    if not data:
        raise ValueError('iterable must be non-empty.')

    def tcat(x):
        return torch.cat(x, dim=dim)

    first = data[0]
    if isinstance(first, torch.Tensor):
        raise ValueError(
            'Use torch.cat instead of delu.cat for concatenating a sequence of tensors.'
            ' Use delu.cat when concatenating a sequence of collections of tensors.'
        )
    elif isinstance(first, tuple):
        constructor = type(first)._make if is_namedtuple(first) else type(first)  # type: ignore  # noqa: E501
        return constructor(tcat([x[i] for x in data]) for i in range(len(first)))
    elif isinstance(first, dict):
        return type(first)((key, tcat([x[key] for x in data])) for key in first)  # type: ignore  # noqa: E501
    elif dataclasses.is_dataclass(first):
        fields = {}
        for field in dataclasses.fields(first):
            if field.type is not torch.Tensor:
                raise ValueError(
                    f'All dataclass fields must be PyTorch Tensors. Found {field.type}'
                )
            fields[field.name] = tcat([getattr(x, field.name) for x in data])
        return type(first)(**fields)
    else:
        raise ValueError(f'The collection type {type(first)} is not supported.')


@deprecated('Instead, use `delu.cat`')
def concat(*args, **kwargs):
    """"""
    return cat(*args, **kwargs)


def iter_batches(
    data: T,
    batch_size: int,
    shuffle: bool = False,
    *,
    generator: Optional[torch.Generator] = None,
    drop_last: bool = False,
) -> Iterator[T]:
    """Iterate over tensor or collection of tensors by (random) batches.

    The function makes batches over the first dimension of the tensors in ``data``
    and returns an iterator over collections of the same type as the input.

    Args:
        data: the tensor or the collection ((named)tuple/dict/dataclass) of tensors.
            If data is a collection, then the tensors must have the same first
            dimension. If data is a dataclass, then all its fields must be tensors.
        batch_size: the batch size. If ``drop_last`` is False, then the last batch can
            be smaller than ``batch_size``.
        shuffle: if True, iterate over random batches (without replacement),
            not sequentially.
        generator: the argument for `torch.randperm` when ``shuffle`` is True.
        drop_last: same as the ``drop_last`` argument for `torch.utils.data.DataLoader`.
            When True and the last batch is smaller then ``batch_size``, then this last
            batch is not returned.
    Returns:
        Iterator over batches.
    Raises:
        ValueError: if the data is empty.

    Note:
        The function lazily indexes to the provided input with batches of indices.
        This works faster than iterating over the tensors in ``data`` with
        `torch.utils.data.DataLoader`.

    See also:
        - `delu.cat`

    Examples:
        .. code-block::

            for epoch in range(n_epochs):
                for batch in delu.iter_batches(data, batch_size, shuffle=True)):
                    ...

        .. testcode::

            a = torch.tensor([0.0, 1.0, 2.0, 3.0, 4.0])
            b = torch.tensor([[0], [10], [20], [30], [40]])
            batch_size = 2

            for batch in delu.iter_batches(a, batch_size):
                assert isinstance(batch, torch.Tensor)
            for batch in delu.iter_batches((a, b), batch_size):
                assert isinstance(batch, tuple) and len(batch) == 2
            for batch in delu.iter_batches({'a': a, 'b': b}, batch_size):
                assert isinstance(batch, dict) and set(batch) == {'a', 'b'}

            from dataclasses import dataclass
            @dataclass
            class Data:
                a: torch.Tensor
                b: torch.Tensor

            for batch in delu.iter_batches(Data(a, b), batch_size):
                assert isinstance(batch, Data)

            ab = delu.cat(delu.iter_batches((a, b), batch_size))
            assert torch.equal(ab[0], a)
            assert torch.equal(ab[1], b)

            n_batches = len(list(delu.iter_batches((a, b), batch_size)))
            assert n_batches == 3
            n_batches = len(list(delu.iter_batches((a, b), batch_size, drop_last=True)))
            assert n_batches == 2
    """
    if not shuffle and generator is not None:
        raise ValueError('When shuffle is False, generator must be None.')

    if isinstance(data, torch.Tensor):
        if not len(data):
            raise ValueError('data must be non-empty')
        item = data
        get_batch = data.__getitem__
    elif isinstance(data, tuple):
        if not data:
            raise ValueError('data must be non-empty')
        item = data[0]
        if any(len(x) != len(item) for x in data):
            raise ValueError('All tensors must have the same first dimension.')
        constructor = type(data)._make if is_namedtuple(data) else type(data)  # type: ignore  # noqa: E501
        get_batch = lambda idx: constructor(x[idx] for x in data)  # type: ignore  # noqa: E731,E501
    elif isinstance(data, dict):
        if not data:
            raise ValueError('data must be non-empty')
        item = next(iter(data.values()))
        if any(len(x) != len(item) for x in data.values()):
            raise ValueError('All tensors must have the same first dimension.')
        get_batch = lambda idx: type(data)({k: v[idx] for k, v in data.items()})  # type: ignore # noqa: E731,E501
    elif dataclasses.is_dataclass(data):
        fields = list(dataclasses.fields(data))
        if not fields:
            raise ValueError('data must be non-empty')
        item = getattr(data, fields[0].name)
        for field in fields:
            if field.type is not torch.Tensor:
                raise ValueError('All dataclass fields must be tensors')
            if len(getattr(data, field.name)) != len(item):
                raise ValueError('All tensors must have the same first dimension.')
        get_batch = lambda idx: type(data)(  # type: ignore  # noqa: E731
            **{field.name: getattr(data, field.name)[idx] for field in fields}
        )
    else:
        raise ValueError(f'The collection {type(data)} is not supported.')

    size = len(item)
    device = item.device
    n_batches = math.ceil(size / batch_size)
    for i, idx in enumerate(
        (
            torch.randperm(size, generator=generator, device=device)
            if shuffle
            else torch.arange(size, device=device)
        ).split(batch_size)
    ):
        if i + 1 == n_batches and len(idx) < batch_size and drop_last:
            return
        yield get_batch(idx)  # type: ignore



import itertools
import math

from pytest import mark, raises

from delu import Stream

from .util import ignore_deprecated_warning


@ignore_deprecated_warning
def test_properties():
    loader = [0]
    stream = Stream(loader)
    assert stream.iteration == 0
    assert stream.epoch == 0
    assert stream.loader is loader
    for attr in 'iteration', 'epoch', 'loader':
        with raises(AttributeError):
            setattr(stream, attr, None)


@ignore_deprecated_warning
def test_bad_loader():
    with raises(AssertionError):
        Stream([])


@ignore_deprecated_warning
def test_increment_epoch():
    stream = Stream(range(3))
    for i in range(10):
        assert stream.epoch == i
        stream.increment_epoch()


@ignore_deprecated_warning
def test_set_loader():
    a = itertools.repeat(0)
    b = itertools.repeat(1)
    stream = Stream([2])
    stream.set_loader(a)
    for x in range(10):
        assert stream.next() == next(b if x % 2 else a)
        stream.set_loader(a if x % 2 else b)


@ignore_deprecated_warning
def test_reload_iterator():
    n = 10
    stream = Stream(range(n))
    for x in range(n):
        assert stream.next() == 0
        stream.reload_iterator()

    stream = Stream(iter(range(n)))
    for x in range(n):
        assert stream.next() == x
        stream.reload_iterator()
    with raises(StopIteration):
        stream.next()


@ignore_deprecated_warning
@mark.parametrize('n', range(1, 5))
def test_next_iteration(n):
    stream = Stream(range(n))
    assert stream.iteration == 0

    for epoch in range(2):
        for x in range(n):
            assert stream.next() == x
            assert stream.iteration == x + epoch * n + 1

    stream = Stream(iter(range(n)))
    for _ in range(n):
        stream.next()
    with raises(StopIteration):
        stream.next()
    with raises(StopIteration):
        stream.next()


@ignore_deprecated_warning
@mark.parametrize('n', range(1, 6))
def test_data(n):
    # iterables that are not iterators
    n_epochs = 10
    stream = Stream(range(n))
    assert [list(stream.data(0)) for _ in range(n_epochs)] == [[]] * n_epochs

    for epoch_size in [None, *list(range(1, 2 * n))]:
        effective_epoch_size = n if epoch_size is None else epoch_size
        max_iteration = n_epochs * effective_epoch_size
        stream = Stream(range(n))
        actual = []
        while stream.iteration < max_iteration:
            actual.append(list(stream.data(epoch_size)))
        assert stream.iteration == max_iteration
        flat_correct = [x % n for x in range(max_iteration)]
        correct = [
            flat_correct[i * effective_epoch_size : (i + 1) * effective_epoch_size]
            for i in range(n_epochs)
        ]
        assert actual == correct

    # count=inf
    stream = Stream(range(n))
    assert [x for _, x in zip(range(1000), stream.data(math.inf))] == [
        x % n for x in range(1000)
    ]

    # infinite iterators
    stream = Stream(itertools.count())
    with raises(ValueError):
        stream.data()
    n_epochs = 10
    for epoch_size in range(1, n):
        stream = Stream(itertools.count())
        actual = [list(stream.data(epoch_size)) for _ in range(n_epochs)]
        flat_correct = list(range(n_epochs * epoch_size))
        correct = [
            flat_correct[i * epoch_size : (i + 1) * epoch_size] for i in range(n_epochs)
        ]
        assert actual == correct

    # finite iterators
    stream = Stream(iter(range(n)))
    with raises(ValueError):
        stream.data()
    n_epochs = 10
    for epoch_size in range(1, 2 * n):
        stream = Stream(iter(range(n)))
        actual = []
        for _ in range(n_epochs):
            actual.append(list(stream.data(epoch_size)))
        flat_correct = list(range(n))
        correct = [
            flat_correct[i * epoch_size : (i + 1) * epoch_size] for i in range(n_epochs)
        ]
        assert actual == correct

    # bad input
    with raises(AssertionError):
        stream.data(10.0)


@ignore_deprecated_warning
def test_epochs():
    stream = Stream(range(3))
    with raises(AssertionError):
        next(stream.epochs(1.0, progress_bar_config=None))
    correct = [0, 1, 2]
    for epoch in stream.epochs(2, progress_bar_config=None):
        assert list(epoch) == correct
    correct = [[0, 1], [2, 0], [1, 2]]
    for i, epoch in enumerate(stream.epochs(2, progress_bar_config=None)):
        assert list(epoch) == correct[i]
    for i, epoch in zip(
        enumerate(stream.epochs(math.inf, progress_bar_config=None)), range(1000)
    ):
        pass
    for (i, epoch), _ in zip(enumerate(stream.epochs(math.inf, math.inf)), range(10)):
        for (j, _), _ in zip(enumerate(epoch), range(10)):
            pass
        assert j == 9
    assert i == 9


@ignore_deprecated_warning
def test_state_dict():
    stream = Stream(range(10))
    stream.next()
    stream.increment_epoch()
    assert stream.state_dict() == {'epoch': 1, 'iteration': 1}

    new_stream = Stream(range(10))
    new_stream.load_state_dict(stream.state_dict())
    assert new_stream.state_dict() == {'epoch': 1, 'iteration': 1}
    assert new_stream.next() == 0
    assert new_stream.state_dict() == {'epoch': 1, 'iteration': 2}
    assert new_stream.next() == 1

    assert stream.next() == 1



# Do Ctrl+F "delu" to see how DeLU is used.
# Run `python mnist.py --help` to see the documentation

# ruff: noqa: F821,E501

import shutil
import subprocess
import sys
from argparse import ArgumentParser
from pathlib import Path

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader, random_split

try:
    from torchvision.datasets import MNIST
    from torchvision.transforms import ToTensor
except ImportError:
    print(
        'Please, install the following packages to proceed:\n'
        '- torchvision (see https://pytorch.org/get-started)'
    )
    raise

import delu


def download_mnist():
    path = Path('MNIST.tar.gz')
    if path.exists():
        path.unlink()
        shutil.rmtree('MNIST', ignore_errors=True)
    subprocess.run(['wget', 'www.di.ens.fr/~lelarge/MNIST.tar.gz'], check=True)
    subprocess.run(['tar', '-zxvf', 'MNIST.tar.gz'], check=True)


def get_dataset(train):
    return MNIST('.', train=train, transform=lambda x: ToTensor()(x).view(-1))


def split_dataset(dataset, ratio):
    size = len(dataset)
    first_size = int(ratio * size)
    return random_split(dataset, [first_size, size - first_size])  # type: ignore[code]


def parse_args():
    parser = ArgumentParser(epilog='Example: python mnist.py')
    parser.add_argument('-d', '--device', default='cpu', type=torch.device)
    parser.add_argument('-n', '--n-epochs', type=int, default=20)
    parser.add_argument('-p', '--early-stopping-patience', type=int, default=1)
    parser.add_argument('-s', '--seed', type=int, default=0)
    parser.add_argument('-c', '--from-checkpoint')
    parser.add_argument('--skip-download', action='store_true')
    return parser.parse_args()


def main():
    assert str(Path.cwd().absolute().resolve()).endswith(
        'delu/examples'
    ), 'Run this script from the "examples" directory'
    args = parse_args()
    assert args.n_epochs > 0

    if not args.skip_download:
        download_mnist()
    # Set seeds in all "standard" libraries: random, numpy, torch
    delu.random.seed(args.seed)
    model = nn.Linear(784, 10).to(args.device)
    optimizer = torch.optim.SGD(model.parameters(), 0.005, 0.9)

    def step(batch):
        # Change device and/or dtype for a collection of tensors.
        X, y = delu.to(batch, args.device)
        return model(X), y

    @torch.inference_mode()
    def evaluate(loader):
        model.eval()
        # Concatenate a sequence of tuples `(batch_logits, batch_y)` into a single tuple.
        logits, y = delu.cat(map(step, loader))
        y_pred = torch.argmax(logits, dim=1).to(y)
        return (y_pred == y).int().sum().item() / len(y)

    train_dataset, val_dataset = split_dataset(get_dataset(True), 0.8)
    test_dataset = get_dataset(False)
    train_loader = DataLoader(train_dataset, batch_size=64, shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=8096)
    test_loader = DataLoader(test_dataset, batch_size=8096)

    # Create a timer, see the usage below.
    timer = delu.Timer()
    # Create a progress tracker for tracking the best score and for early stopping.
    progress = delu.ProgressTracker(args.early_stopping_patience, 0.005)
    checkpoint_path = 'checkpoint.pt'
    if args.from_checkpoint:
        assert args.from_checkpoint != 'checkpoint.pt'
        assert Path(args.from_checkpoint).exists()
        checkpoint = torch.load(args.from_checkpoint)
        start_epoch = checkpoint['epoch'] + 1
        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        timer = checkpoint['timer']
        progress = checkpoint['progress']
        # Restore the global random state for full reproducibility.
        delu.random.set_state(checkpoint['random_state'])
        print('Resuming from the checkpoint.\n')
    else:
        start_epoch = 0

    for epoch in range(start_epoch, args.n_epochs):
        print(f'\nEpoch {epoch} started')

        with timer:
            model.train()
            for batch in train_loader:
                optimizer.zero_grad()
                F.cross_entropy(*step(batch)).backward()
                optimizer.step()

        accuracy = evaluate(val_loader)
        progress.update(accuracy)
        if progress.success:
            print('New best score!')
            torch.save(
                {
                    'epoch': epoch,
                    'model': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                    'timer': timer,
                    'progress': progress,
                    # Save the global random state.
                    'random_state': delu.random.get_state(),
                },
                checkpoint_path,
            )

        msg = (
            f'Epoch {epoch} finished. '
            f'Time elapsed: {timer}. '
            f'Validation accuracy: {accuracy:.4f}.'
        )
        print(msg)
        if progress.fail:
            break

    timer.pause()
    model.load_state_dict(torch.load(checkpoint_path)['model'])
    print(
        f'\nTraining stopped after the epoch {epoch}.\n'
        f'Total training time: {timer}\n'
        f'Test accuracy: {evaluate(test_loader)}'
    )

    print('Freeing memory (for fun, not for profit) ...')
    del model, optimizer, step, evaluate
    # A handy function for freeing GPU memory.
    delu.cuda.free_memory()
    print('\nDONE.')
    if not args.from_checkpoint:
        # TODO replace `join` with `shlex.join` in 3.8
        print(
            '\nNow, you can test if the last epochs can be reproduced when the '
            'training is resumed from the last available checkpoint. For that, run:\n'
            'cp checkpoint.pt a.pt\n'
            f'python {" ".join(sys.argv)} -c a.pt'
        )


if __name__ == '__main__':
    main()



"""An addition to `torch.cuda`."""
import gc

import torch


def free_memory() -> None:
    """Free GPU memory: `torch.cuda.synchronize` + `gc.collect` + `torch.cuda.empty_cache`.

    Warning:
        There is a small chunk of GPU-memory (occupied by drivers) that is impossible to
        free. It is a `torch` "limitation", so the function inherits this property.

    Example:
        .. testcode::

            delu.cuda.free_memory()
    """  # noqa: E501
    # Step 1: finish the ongoing computations.
    if torch.cuda.is_available():
        # torch has wrong .pyi
        torch.cuda.synchronize()  # type: ignore
    # Step 2: collect unused objects.
    gc.collect()
    # Step 3: free GPU-cache.
    if torch.cuda.is_available():
        torch.cuda.empty_cache()



"""An addition to `torch.nn`."""

from typing import Callable

import torch.nn


class Lambda(torch.nn.Module):
    """A parameter-free module for wrapping callables.

    Examples:
        .. testcode::

            module = delu.nn.Lambda(torch.square)
            assert torch.equal(module(torch.tensor(3)), torch.tensor(9))

            # Any callable can be wrapped in Lambda:
            module = delu.nn.Lambda(lambda x, y, z: x + y + z)
            assert module(1, 2, z=3) == 6
    """

    def __init__(self, fn: Callable):
        """Initialize self."""
        super().__init__()
        self.fn = fn

    def forward(self, *args, **kwargs):
        """Perform the forward pass."""
        return self.fn(*args, **kwargs)



import random

import numpy as np
import pytest
import torch

import delu

from .util import ignore_deprecated_warning


@ignore_deprecated_warning
@pytest.mark.parametrize('train', [False, True])
@pytest.mark.parametrize('grad', [False, True])
@pytest.mark.parametrize('n_models', range(3))
def test_evaluation(train, grad, n_models):
    if not n_models:
        with pytest.raises(AssertionError):
            with delu.evaluation():
                pass
        return

    torch.set_grad_enabled(grad)
    models = [torch.nn.Linear(1, 1) for _ in range(n_models)]
    for x in models:
        x.train(train)
    with delu.evaluation(*models):
        assert all(not x.training for x in models[:-1])
        assert not torch.is_grad_enabled()
    assert torch.is_grad_enabled() == grad
    for x in models:
        x.train(train)

    @delu.evaluation(*models)
    def f():
        assert all(not x.training for x in models[:-1])
        assert not torch.is_grad_enabled()
        for x in models:
            x.train(train)

    for _ in range(3):
        f()
        assert torch.is_grad_enabled() == grad


@ignore_deprecated_warning
def test_evaluation_generator():
    with pytest.raises(AssertionError):

        @delu.evaluation(torch.nn.Linear(1, 1))
        def generator():
            yield 1


@ignore_deprecated_warning
def test_improve_reproducibility():
    def f():
        upper_bound = 100
        return [
            random.randint(0, upper_bound),
            np.random.randint(upper_bound),
            torch.randint(upper_bound, (1,))[0].item(),
        ]

    for seed in [None, 0, 1, 2]:
        seed = delu.improve_reproducibility(seed)
        assert not torch.backends.cudnn.benchmark
        assert torch.backends.cudnn.deterministic
        results = f()
        delu.random.seed(seed)
        assert results == f()



import pickle
from time import perf_counter, sleep

import pytest

import delu


def test_progress_tracker():
    score = -999999999

    # test initial state
    tracker = delu.ProgressTracker(0)
    assert not tracker.success
    assert not tracker.fail

    # test successful update
    tracker.update(score)
    assert tracker.best_score == score
    assert tracker.success

    # test failed update
    tracker.update(score)
    assert tracker.best_score == score
    assert tracker.fail

    # test forget_bad_updates, reset
    tracker.forget_bad_updates()
    assert tracker.best_score == score
    tracker.reset()
    assert tracker.best_score is None
    assert not tracker.success and not tracker.fail

    # test positive patience
    tracker = delu.ProgressTracker(1)
    tracker.update(score - 1)
    assert tracker.success
    tracker.update(score)
    assert tracker.success
    tracker.update(score)
    assert not tracker.success and not tracker.fail
    tracker.update(score)
    assert tracker.fail

    # test positive min_delta
    tracker = delu.ProgressTracker(0, 2)
    tracker.update(score - 2)
    assert tracker.success
    tracker.update(score)
    assert tracker.fail
    tracker.reset()
    tracker.update(score - 3)
    tracker.update(score)
    assert tracker.success

    # patience=None
    tracker = delu.ProgressTracker(None)
    for i in range(100):
        tracker.update(-i)
        assert not tracker.fail


def test_timer():
    with pytest.raises(AssertionError):
        delu.Timer().pause()

    # initial state, run
    timer = delu.Timer()
    sleep(0.001)
    assert not timer()
    timer.run()
    assert timer()

    # pause
    timer.pause()
    timer.pause()  # two pauses in a row
    x = timer()
    sleep(0.001)
    assert timer() == x

    # run
    timer.pause()
    x = timer()
    timer.run()
    timer.run()  # two runs in a row
    assert timer() != x
    timer.pause()
    x = timer()
    sleep(0.001)
    assert timer() == x
    timer.run()

    # reset
    timer.reset()
    assert not timer()


def test_timer_measurements():
    x = perf_counter()
    sleep(0.1)
    correct = perf_counter() - x
    timer = delu.Timer()
    timer.run()
    sleep(0.1)
    actual = timer()
    # the allowed deviation was obtained from manual runs on my laptop so the test may
    # behave differently on other hardware
    assert actual == pytest.approx(correct, abs=0.01)


def test_timer_context():
    with delu.Timer() as timer:
        sleep(0.01)
    assert timer() > 0.01
    assert timer() == timer()

    timer = delu.Timer()
    timer.run()
    sleep(0.01)
    timer.pause()
    with timer:
        sleep(0.01)
    assert timer() > 0.02
    assert timer() == timer()


def test_timer_pickle():
    timer = delu.Timer()
    timer.run()
    sleep(0.01)
    timer.pause()
    value = timer()
    sleep(0.01)
    assert pickle.loads(pickle.dumps(timer))() == timer() == value


def test_timer_format():
    def make_timer(x):
        timer = delu.Timer()
        timer._shift = x
        return timer

    assert str(make_timer(1)) == '0:00:01'
    assert str(make_timer(1.1)) == '0:00:01'
    assert make_timer(7321).format('%Hh %Mm %Ss') == '02h 02m 01s'



import inspect
import secrets
from contextlib import ContextDecorator
from typing import Any, Optional

import torch
import torch.nn as nn

from . import random as delu_random
from ._utils import deprecated


@deprecated(
    'Instead, use `delu.random.seed` and manually set flags mentioned'
    ' in the `PyTorch docs on reproducibility <https://pytorch.org/docs/stable/notes/randomness.html>`_'  # noqa: E501
)
def improve_reproducibility(
    base_seed: Optional[int], one_cuda_seed: bool = False
) -> int:
    """Set seeds and turn off non-deterministic algorithms.

    Do everything possible to improve reproducibility for code that relies on global
    random number generators. See also the note below.

    Sets:

    1. seeds in `random`, `numpy.random`, `torch`, `torch.cuda`
    2. `torch.backends.cudnn.benchmark` to `False`
    3. `torch.backends.cudnn.deterministic` to `True`

    Args:
        base_seed: the argument for `delu.random.seed`. If `None`, a high-quality base
            seed is generated instead.
        one_cuda_seed: the argument for `delu.random.seed`.

    Returns:
        base_seed: if ``base_seed`` is set to `None`, the generated base seed is
            returned; otherwise, ``base_seed`` is returned as is

    Note:
        If you don't want to choose the base seed, but still want to have a chance to
        reproduce things, you can use the following pattern::

            print('Seed:', delu.improve_reproducibility(None))

    Note:
        100% reproducibility is not always possible in PyTorch. See
        `this page <https://pytorch.org/docs/stable/notes/randomness.html>`_ for
        details.

    Examples:
        .. testcode::

            assert delu.improve_reproducibility(0) == 0
            seed = delu.improve_reproducibility(None)
    """
    torch.backends.cudnn.benchmark = False  # type: ignore
    torch.backends.cudnn.deterministic = True  # type: ignore
    if base_seed is None:
        # See https://numpy.org/doc/1.18/reference/random/bit_generators/index.html#seeding-and-entropy  # noqa
        base_seed = secrets.randbits(128) % (2**32 - 1024)
    else:
        assert base_seed < (2**32 - 1024)
    delu_random.seed(base_seed, one_cuda_seed)
    return base_seed


@deprecated('Instead, use ``model.eval()`` + ``torch.no_inference/no_grad``')
class evaluation(ContextDecorator):
    """Context-manager & decorator for models evaluation.

    This code... ::

        with delu.evaluation(model):  # or: evaluation(model_0, model_1, ...)
            ...
        @delu.evaluation(model)  # or: @evaluation(model_0, model_1, ...)
        def f():
            ...

    ...is equivalent to the following ::

        context = getattr(torch, 'inference_mode', torch.no_grad)
        with context():
            model.eval()
            ...
        @context()
        def f():
            model.eval()
            ...

    Args:
        modules
    Note:
        The training status of modules is undefined once a context is finished or a
        decorated function returns.
    Warning:
        The function must be used in the same way as `torch.no_grad` and
        `torch.inference_mode`, i.e. only as a context manager or a decorator as shown
        below in the examples. Otherwise, the behaviour is undefined.
    Warning:
        Contrary to `torch.no_grad` and `torch.inference_mode`, the function cannot be
        used to decorate generators. So, in the case of generators, you have to manually
        create a context::

            def my_generator():
                with delu.evaluation(...):
                    for a in b:
                        yield c

    Examples:

        .. testcode::

            a = torch.nn.Linear(1, 1)
            b = torch.nn.Linear(2, 2)
            with delu.evaluation(a):
                ...
            with delu.evaluation(a, b):
                ...
            @delu.evaluation(a)
            def f():
                ...
            @delu.evaluation(a, b)
            def f():
                ...
    """

    def __init__(self, *modules: nn.Module) -> None:
        assert modules
        self._modules = modules
        self._torch_context: Any = None

    def __call__(self, func):
        """Decorate a function with an evaluation context.

        Args:
            func
        Raises:
            AssertionError: if :code:`func` is a generator
        """
        assert not inspect.isgeneratorfunction(
            func
        ), f'{self.__class__} cannot be used to decorate generators.'
        return super().__call__(func)

    def __enter__(self) -> None:
        assert self._torch_context is None
        self._torch_context = getattr(torch, 'inference_mode', torch.no_grad)()
        self._torch_context.__enter__()  # type: ignore
        for m in self._modules:
            m.eval()

    def __exit__(self, *exc):
        assert self._torch_context is not None
        result = self._torch_context.__exit__(*exc)  # type: ignore
        self._torch_context = None
        return result



"""Random sampling utilities."""

import random
from contextlib import contextmanager
from typing import Any, Dict

import numpy as np
import torch


def seed(base_seed: int, one_cuda_seed: bool = False) -> None:
    """Set seeds in `random`, `numpy` and `torch`.

    For all libraries, different seeds (which are *deterministically* computed based
    on the ``base_seed`` argument) are set (see the note below).

    Args:
        base_seed: the number used to determine random seeds for all libraries and
            hardware. Must be a non-negative number less than ``2 ** 32 - 10000``.
        one_cuda_seed: if `True`, then the same seed will be set for all cuda devices,
            otherwise, different seeds will be set for all cuda devices.
    Raises:
        AssertionError: if the seed is not within the required interval

    Note:
        Different seeds are set to avoid situations where different libraries or devices
        generate the same random sequences. See
        `this comment <https://github.com/PyTorchLightning/pytorch-lightning/pull/6960#issuecomment-818393659>`_
        for details.

    Examples:
        .. testcode::

            delu.random.seed(0)
    """  # noqa: E501
    assert 0 <= base_seed < 2**32 - 10000
    random.seed(base_seed)
    np.random.seed(base_seed + 1)
    torch.manual_seed(base_seed + 2)
    cuda_seed = base_seed + 3
    if one_cuda_seed:
        torch.cuda.manual_seed_all(cuda_seed)
    elif torch.cuda.is_available():
        # the following check should never succeed since torch.manual_seed also calls
        # torch.cuda.manual_seed_all() inside; but let's keep it just in case
        if not torch.cuda.is_initialized():
            torch.cuda.init()
        # Source: https://github.com/pytorch/pytorch/blob/2f68878a055d7f1064dded1afac05bb2cb11548f/torch/cuda/random.py#L109
        for i in range(torch.cuda.device_count()):
            default_generator = torch.cuda.default_generators[i]
            default_generator.manual_seed(cuda_seed + i)


def get_state() -> Dict[str, Any]:
    """Aggregate global random states from `random`, `numpy` and `torch`.

    The function is useful for creating checkpoints and allows resuming activities
    dependent on global random states (e.g. data streams) in a reproducible manner.
    Also, see the note below.
    The result of this function can be passed to `set_state`.

    Returns:
        state

    See also:
        `delu.random.set_state`

    Example::

        # Resuming from a checkpoint:
        checkpoint_path = ...
        if checkpoint_path.exists():
            checkpoint = torch.load(checkpoint_path)
            delu.random.set_state(checkpoint['random_state'])
        ...
        # Training:
        for batch in batches:
            ...
            if step % checkpoint_frequency == 0:
                torch.save(
                    {
                        'model': ...,
                        'optimizer': ...,
                        'random_state': delu.random.get_state(),
                    },
                    checkpoint_path,
                )
    """
    return {
        'random': random.getstate(),
        'numpy.random': np.random.get_state(),
        'torch.random': torch.random.get_rng_state(),
        'torch.cuda': torch.cuda.get_rng_state_all(),  # type: ignore
    }


def set_state(state: Dict[str, Any]) -> None:
    """Set global random states in `random`, `numpy` and `torch`.

    See `get_state` for a usage example.

    Args:
        state: the global RNG states. Must be produced by `get_state`. The size of the
            list ``state['torch.cuda']`` must be equal to the number of available cuda
            devices.

    See also:
        `delu.random.get_state`

    Raises:
        AssertionError: if ``torch.cuda.device_count() != len(state['torch.cuda'])``
    """
    random.setstate(state['random'])
    np.random.set_state(state['numpy.random'])
    torch.random.set_rng_state(state['torch.random'])
    assert torch.cuda.device_count() == len(state['torch.cuda'])
    torch.cuda.set_rng_state_all(state['torch.cuda'])  # type: ignore


@contextmanager
def preserve_state():
    """Decorator and context manager for preserving global random state.

    The function saves the global random state when entering the context
    and restores it on exit.

    Examples:
        .. testcode::

            import random

            f = lambda: (
                random.randint(0, 10),
                np.random.randint(10),
                torch.randint(10, (1,)).item()
            )
            with delu.random.preserve_state():
                a = f()
            b = f()
            assert a == b

            @delu.random.preserve_state()
            def g():
                return f()

            a = g()
            b = f()
            assert a == b

    """
    state = get_state()
    try:
        yield
    finally:
        set_state(state)



import pytest
import torch
from torch.utils.data import DataLoader, TensorDataset

import delu.data

from .util import ignore_deprecated_warning


def test_enumerate():
    dataset = TensorDataset(torch.arange(10), torch.arange(10))
    x = delu.data.Enumerate(dataset)
    assert x.dataset is dataset
    assert len(x) == 10
    assert x[3] == (3, (torch.tensor(3), torch.tensor(3)))


def test_fndataset():
    dataset = delu.data.FnDataset(lambda x: x * 2, 3)
    assert len(dataset) == 3
    assert dataset[0] == 0
    assert dataset[1] == 2
    assert dataset[2] == 4

    dataset = delu.data.FnDataset(lambda x: x * 2, 3, lambda x: x * 3)
    assert len(dataset) == 3
    assert dataset[0] == 0
    assert dataset[1] == 6
    assert dataset[2] == 12

    dataset = delu.data.FnDataset(lambda x: x * 2, [1, 10, 100])
    assert len(dataset) == 3
    assert dataset[0] == 2
    assert dataset[1] == 20
    assert dataset[2] == 200

    dataset = delu.data.FnDataset(lambda x: x * 2, (x for x in range(0, 10, 4)))
    assert len(dataset) == 3
    assert dataset[0] == 0
    assert dataset[1] == 8
    assert dataset[2] == 16


def test_index_dataset():
    n = 10
    d = delu.data.IndexDataset(n)
    assert len(d) == n
    for i in range(n):
        assert d[i] == i
    with pytest.raises(IndexError):
        d[-1]
    with pytest.raises(IndexError):
        d[n]


@ignore_deprecated_warning
def test_collate():
    # just test that the function is still a valid alias
    assert torch.equal(delu.data.collate([1])[0], torch.tensor(1))


@ignore_deprecated_warning
def test_iloader():
    with pytest.raises(AssertionError):
        delu.data.IndexLoader(0)

    for x in range(1, 10):
        assert len(delu.data.IndexLoader(x)) == x

    data = torch.arange(10)
    for batch_size in range(1, len(data) + 1):
        torch.manual_seed(batch_size)
        correct = list(DataLoader(data, batch_size, shuffle=True, drop_last=True))
        torch.manual_seed(batch_size)
        actual = list(
            delu.data.IndexLoader(len(data), batch_size, shuffle=True, drop_last=True)
        )
        for x, y in zip(actual, correct):
            assert torch.equal(x, y)



"""Tools related to devices, memory, etc."""

from typing import Any, Dict

from ._utils import deprecated
from .cuda import free_memory as free_memory_original

try:
    import pynvml
except ImportError:
    pynvml = None
if pynvml is not None:
    from pynvml import NVMLError_LibraryNotFound


def _to_str(x):
    if isinstance(x, str):
        return x
    elif isinstance(x, bytes):
        return str(x, 'utf-8')
    else:
        raise ValueError('Internal error')


@deprecated('Instead, use `delu.cuda.free_memory`.')
def free_memory(*args, **kwargs) -> None:
    """"""
    return free_memory_original(*args, **kwargs)


@deprecated('Instead, use functions from `torch.cuda`')
def get_gpus_info() -> Dict[str, Any]:
    """Get information about GPU devices: driver version, memory, utilization etc.

    The example below shows what kind of information is returned as the result. All
    figures about memory are given in bytes.

    Returns:
        Information about GPU devices.
    Raises:
        RuntimeError: if necessary cuda-related libraries are not found. Usually, it
            means that the function is run on a machine without GPU.

    Warning:
        The 'devices' value contains information about *all* gpus regardless of the
        value of ``CUDA_VISIBLE_DEVICES``.

    Examples:
        .. code-block::

            print(delu.hardware.get_gpu_info())

        Output example (formatted for convenience):

        .. code-block:: none

            {
                'driver': '440.33.01',
                'devices': [
                    {
                        'name': 'GeForce RTX 2080 Ti',
                        'memory_total': 11554717696,
                        'memory_free': 11554652160,
                        'memory_used': 65536,
                        'utilization': 0,
                    },
                    {
                        'name': 'GeForce RTX 2080 Ti',
                        'memory_total': 11552096256,
                        'memory_free': 11552030720,
                        'memory_used': 65536,
                        'utilization': 0,
                    },
                ],
            }
    """
    if pynvml is None:
        raise RuntimeError(
            'To use this function, install pynvml via `pip install pynvml<12.0`'
        )

    try:
        pynvml.nvmlInit()
    except NVMLError_LibraryNotFound as err:
        raise RuntimeError(
            'Failed to get information about GPU memory. '
            'Make sure that you actually have GPU and all relevant software installed.'
        ) from err
    n_devices = pynvml.nvmlDeviceGetCount()
    devices = []
    for device_id in range(n_devices):
        handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
        memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        devices.append(
            {
                'name': _to_str(pynvml.nvmlDeviceGetName(handle)),
                'memory_total': memory_info.total,
                'memory_free': memory_info.free,
                'memory_used': memory_info.used,
                'utilization': pynvml.nvmlDeviceGetUtilizationRates(handle).gpu,
            }
        )
    return {
        'driver': _to_str(pynvml.nvmlSystemGetDriverVersion()),
        'devices': devices,
    }



"""An addition to `torch.utils.data`."""

from typing import Any, Callable, Iterable, Optional, Tuple, TypeVar, Union

import torch
from torch.utils.data import DataLoader, Dataset

from ._stream import Stream  # noqa: F401
from ._utils import deprecated

T = TypeVar('T')


class Enumerate(Dataset):
    """Make dataset return both indices and items.

    .. rubric:: Tutorial

    .. testcode::

        from torch.utils.data import DataLoader, TensorDataset
        X, y = torch.randn(9, 2), torch.randn(9)
        dataset = TensorDataset(X, y)
        for batch_idx, batch in DataLoader(delu.data.Enumerate(dataset), batch_size=3):
            print(batch_idx)

    .. testoutput::

        tensor([0, 1, 2])
        tensor([3, 4, 5])
        tensor([6, 7, 8])
    """

    def __init__(self, dataset: Dataset) -> None:
        """Initialize self.

        Args:
            dataset
        """
        self._dataset = dataset

    @property
    def dataset(self) -> Dataset:
        """Access the underlying dataset.

        Returns:
            The dataset.
        """
        return self._dataset

    def __len__(self) -> int:
        """Get the length of the underlying dataset."""
        return len(self._dataset)  # type: ignore

    def __getitem__(self, index) -> Tuple[Any, Any]:
        """Return index and the corresponding item from the underlying dataset.

        Args:
            index
        Returns:
            (index, item)
        """
        return index, self._dataset[index]


class FnDataset(Dataset):
    """Create simple PyTorch datasets without classes and inheritance.

    `FnDataset` allows avoiding implementing `~torch.utils.data.Dataset` classes in
    simple cases.

    .. rubric:: Tutorial

    First, a quick example. Without `FnDataset`::

        from PIL import Image

        class ImagesList(Dataset):
            def __init__(self, filenames, transform):
                self.filenames = filenames
                self.transform = transform

            def __len__(self):
                return len(self.filenames)

            def __getitem__(self, index):
                return self.transform(Image.open(self.filenames[index]))

        dataset = ImagesList(filenames, transform)

    With `FnDataset`::

        dataset = delu.data.FnDataset(Image.open, filenames, transform)
        # Cache images after the first load:
        from functools import lru_cache
        dataset = delu.data.FnDataset(lru_cache(None)(Image.open), filenames)

    In other words, with the vanilla PyTorch, in order to create a dataset,
    you have to inherit from `torch.utils.data.Dataset` and implement three methods:

    - ``__init__``
    - ``__len__``
    - ``__getitem__``

    With `FnDataset` the only thing you *may* need to implement is the ``fn``
    argument that will power ``__getitem__``. The easiest way to learn
    `FnDataset` is to go through the examples below.

    A list of images::

        dataset = delu.data.FnDataset(Image.open, filenames)
        # dataset[i] returns Image.open(filenames[i])

    A list of images that are cached after the first load::

        from functools import lru_cache
        dataset = delu.data.FnDataset(lru_cache(None)(Image.open), filenames)

    `pathlib.Path` is handy for creating datasets that read from files::

        images_dir = Path(...)
        dataset = delu.data.FnDataset(Image.open, images_dir.iterdir())

    If you only need files with specific extensions::

        dataset = delu.data.FnDataset(Image.open, images_dir.glob('*.png'))

    If you only need files with specific extensions located in all subfolders::

        dataset = delu.data.FnDataset(
            Image.open, (x for x in images_dir.rglob('**/*.png') if condition(x))
        )

    A segmentation dataset::

        image_filenames = ...
        gt_filenames = ...

        def get(i):
            return Image.open(image_filenames[i]), Image.open(gt_filenames[i])

        dataset = delu.data.FnDataset(get, len(image_filenames))

    A dummy dataset that demonstrates that `FnDataset` is a very general thing:

    .. testcode::

        def f(x):
            return x * 10

        def g(x):
            return x * 2

        dataset = delu.data.FnDataset(f, 3, g)
        # dataset[i] returns g(f(i))
        assert len(dataset) == 3
        assert dataset[0] == 0
        assert dataset[1] == 20
        assert dataset[2] == 40

    """

    def __init__(
        self,
        fn: Callable[..., T],
        args: Union[int, Iterable],
        transform: Optional[Callable[[T], Any]] = None,
    ) -> None:
        """Initialize self.

        Args:
            fn: the function that produces values based on arguments from ``args``
            args: arguments for ``fn``. If an iterable, but not a list, then is
                casted to a list. If an integer, then the behavior is the same as for
                ``list(range(args))``. The size of ``args`` defines the return
                value for `FnDataset.__len__`.
            transform: if presented, is applied to the return value of `fn` in
                `FnDataset.__getitem__`

        Examples:
            .. code-block::

                import PIL.Image as Image
                import torchvision.transforms as T

                dataset = delu.data.FnDataset(Image.open, filenames, T.ToTensor())
        """
        self._fn = fn
        if isinstance(args, Iterable):
            if not isinstance(args, list):
                args = list(args)
        self._args = args
        self._transform = transform

    def __len__(self) -> int:
        """Get the dataset size.

        See `FnDataset` for details.

        Returns:
            size
        """
        return len(self._args) if isinstance(self._args, list) else self._args

    def __getitem__(self, index: int) -> Any:
        """Get value by index.

        See `FnDataset` for details.

        Args:
            index
        Returns:
            value
        Raises:
            IndexError: if ``index >= len(self)``
        """
        if isinstance(self._args, list):
            x = self._args[index]
        elif index < self._args:
            x = index
        else:
            raise IndexError(f'Index {index} is out of range')
        x = self._fn(x)
        return x if self._transform is None else self._transform(x)


class IndexDataset(Dataset):
    """A trivial dataset that yeilds indices back to user (useful for DDP).

    This simple dataset is useful when:

    1. you need a dataloader that yeilds batches of *indices* instead of *objects*
    2. AND you work in the `Distributed Data Parallel <https://pytorch.org/tutorials/intermediate/ddp_tutorial.html>`_ setup

    Note:
        If only the first condition is true, consider using the combinatation of
        `torch.randperm` and `torch.Tensor.split` instead.

    Example::

        from torch.utils.data import DataLoader
        from torch.utils.data.distributed import DistributedSampler

        train_size = 123456
        batch_size = 123
        dataset = delu.data.IndexDataset(dataset_size)
        for i in range(train_size):
            assert dataset[i] == i
        dataloader = DataLoader(
            dataset,
            batch_size,
            sampler=DistributedSampler(dataset)
        )

        for epoch in range(n_epochs):
            for batch_indices in dataloader:
                ...
    """  # noqa: E501

    def __init__(self, size: int) -> None:
        """Initialize self.

        Args:
            size: the dataset size
        """
        if size < 1:
            raise ValueError('size must be positive')
        self.size = size

    def __len__(self) -> int:
        """Get the dataset size."""
        return self.size

    def __getitem__(self, i: int) -> int:
        """Get the same index back.

        The index must be an integer from ``range(len(self))``.
        """
        if i < 0 or i >= self.size:
            raise IndexError(
                f"index {i} is out of range (dataset's size is {self.size})"
            )
        return i


@deprecated('Instead, use `delu.data.IndexDataset` and `~torch.utils.data.DataLoader`')
def make_index_dataloader(size: int, *args, **kwargs) -> DataLoader:
    """Make `~torch.utils.data.DataLoader` over indices instead of data.

    This is just a shortcut for
    ``torch.utils.data.DataLoader(delu.data.IndexDataset(...), ...)``.

    Args:
        size: the dataset size
        *args: positional arguments for `torch.utils.data.DataLoader`
        **kwargs: keyword arguments for `torch.utils.data.DataLoader`
    Raises:
        ValueError: for invalid inputs
    Examples:

        Usage for training:

        .. code-block::

            train_loader = delu.data.make_index_dataloader(
                len(train_dataset), batch_size, shuffle=True
            )
            for epoch in range(n_epochs):
                for i_batch in train_loader:
                    x_batch = X[i_batch]
                    y_batch = Y[i_batch]
                    ...

        Other examples:

        .. testcode::

            dataset_size = 10  # len(dataset)
            for batch_idx in delu.data.make_index_dataloader(
                dataset_size, batch_size=3
            ):
                print(batch_idx)

        .. testoutput::

            tensor([0, 1, 2])
            tensor([3, 4, 5])
            tensor([6, 7, 8])
            tensor([9])

        .. testcode::

            dataset_size = 10  # len(dataset)
            for batch_idx in delu.data.make_index_dataloader(
                dataset_size, 3, drop_last=True
            ):
                print(batch_idx)

        .. testoutput::

            tensor([0, 1, 2])
            tensor([3, 4, 5])
            tensor([6, 7, 8])
    See also:
        `delu.iter_batches`
    """
    return DataLoader(IndexDataset(size), *args, **kwargs)


@deprecated('Instead, use `delu.data.IndexDataset` and `~torch.utils.data.DataLoader`')
class IndexLoader:
    """Like `~torch.utils.data.DataLoader`, but over indices instead of data.

    **The shuffling logic is delegated to the native PyTorch DataLoader**, i.e. no
    custom logic is performed under the hood. The data loader which actually generates
    indices is available as `IndexLoader.loader`.

    Examples:

        Usage for training:

        .. code-block::

            train_loader = delu.data.IndexLoader(
                len(train_dataset), batch_size, shuffle=True
            )
            for epoch in range(n_epochs):
                for batch_idx in train_loader:
                    ...

        Other examples:

        .. testcode::

            dataset_size = 10  # len(dataset)
            for batch_idx in delu.data.IndexLoader(dataset_size, batch_size=3):
                print(batch_idx)

        .. testoutput::

            tensor([0, 1, 2])
            tensor([3, 4, 5])
            tensor([6, 7, 8])
            tensor([9])

        .. testcode::

            dataset_size = 10  # len(dataset)
            for batch_idx in delu.data.IndexLoader(dataset_size, 3, drop_last=True):
                print(batch_idx)

        .. testoutput::

            tensor([0, 1, 2])
            tensor([3, 4, 5])
            tensor([6, 7, 8])

    See also:
        `delu.iter_batches`
    """

    def __init__(
        self, size: int, *args, device: Union[int, str, torch.device] = 'cpu', **kwargs
    ) -> None:
        """Initialize self.

        Args:
            size: the number of items (for example, :code:`len(dataset)`)
            *args: positional arguments for `torch.utils.data.DataLoader`
            device: if not CPU, then all indices are materialized and moved to the
                device at the beginning of every loop. It can be useful when the indices
                are applied to non-CPU data (e.g. CUDA-tensors) and moving data between
                devices takes non-negligible time (which can happen in the case of
                simple and fast models like MLPs).
            **kwargs: keyword arguments for `torch.utils.data.DataLoader`
        Raises:
            AssertionError: if size is not positive
        """
        assert size > 0
        self._batch_size = args[0] if args else kwargs.get('batch_size', 1)
        self._loader = DataLoader(IndexDataset(size), *args, **kwargs)
        if isinstance(device, (int, str)):
            device = torch.device(device)
        self._device = device

    @property
    def loader(self) -> DataLoader:
        """The underlying DataLoader."""
        return self._loader

    def __len__(self) -> int:
        """Get the size of the underlying DataLoader."""
        return len(self.loader)

    def __iter__(self):
        return iter(
            self._loader
            if self._device.type == 'cpu'
            else torch.cat(list(self.loader)).to(self._device).split(self._batch_size)
        )


@deprecated('Instead, use `torch.utils.data.dataloader.default_collate`')
def collate(iterable: Iterable) -> Any:
    """Almost an alias for :code:`torch.utils.data.dataloader.default_collate`.
    Namely, the input is allowed to be any kind of iterable, not only a list. Firstly,
    if it is not a list, it is transformed to a list. Then, the list is passed to the
    original function and the result is returned as is.
    """
    if not isinstance(iterable, list):
        iterable = list(iterable)
    # > Module has no attribute "default_collate"
    return torch.utils.data.dataloader.default_collate(iterable)  # type: ignore



import datetime
import enum
import time
from typing import Any, Dict, Optional


class _ProgressStatus(enum.Enum):
    NEUTRAL = enum.auto()
    SUCCESS = enum.auto()
    FAIL = enum.auto()


class ProgressTracker:
    """Tracks the best score, helps with early stopping.

    For `~ProgressTracker`, **the greater score is the better score**.
    At any moment the tracker is in one of the following states:

    - *success*: the last update increased the best score
    - *fail*: last ``n > patience`` updates did not improve the best score
    - *neutral*: if neither success nor fail

    .. rubric:: Tutorial

    .. testcode::

        progress = delu.ProgressTracker(2)
        progress.update(-999999999)
        assert progress.success  # the first update always updates the best score

        progress.update(123)
        assert progress.success
        assert progress.best_score == 123

        progress.update(0)
        assert not progress.success and not progress.fail

        progress.update(123)
        assert not progress.success and not progress.fail
        progress.update(123)
        # patience is 2 and the best score is not updated for more than 2 steps
        assert progress.fail
        assert progress.best_score == 123  # fail doesn't affect the best score
        progress.update(123)
        assert progress.fail  # still no improvements

        progress.forget_bad_updates()
        assert not progress.fail and not progress.success
        assert progress.best_score == 123
        progress.update(0)
        assert not progress.fail  # just 1 bad update (the patience is 2)

        progress.reset()
        assert not progress.fail and not progress.success
        assert progress.best_score is None
    """

    def __init__(self, patience: Optional[int], min_delta: float = 0.0) -> None:
        """Initialize self.

        Args:
            patience: Allowed number of unsuccessfull updates. For example, if patience
                is 2, then 2 unsuccessfull updates in a row is not a fail,
                but 3 unsuccessfull updates in a row is a fail.
                `None` means "infinite patience" and the progress tracker is never
                in the "fail" state.
            min_delta: the minimal improvement over the current best score
                to count it as success.

        Examples:
            .. testcode::

                progress = delu.ProgressTracker(2)
                progress = delu.ProgressTracker(3, 0.1)
        """
        self._patience = patience
        self._min_delta = float(min_delta)
        self._best_score: Optional[float] = None
        self._status = _ProgressStatus.NEUTRAL
        self._bad_counter = 0

    @property
    def best_score(self) -> Optional[float]:
        """The best score so far.

        If the tracker is just created/reset, return `None`.
        """
        return self._best_score

    @property
    def success(self) -> bool:
        """Check if the tracker is in the "success" state."""
        return self._status == _ProgressStatus.SUCCESS

    @property
    def fail(self) -> bool:
        """Check if the tracker is in the "fail" state."""
        return self._status == _ProgressStatus.FAIL

    def _set_success(self, score: float) -> None:
        self._best_score = score
        self._status = _ProgressStatus.SUCCESS
        self._bad_counter = 0

    def update(self, score: float) -> None:
        """Submit a new score and update the tracker's state accordingly.

        Args:
            score: the score to use for the update.
        """
        if self._best_score is None:
            self._set_success(score)
        elif score > self._best_score + self._min_delta:
            self._set_success(score)
        else:
            self._bad_counter += 1
            self._status = (
                _ProgressStatus.FAIL
                if self._patience is not None and self._bad_counter > self._patience
                else _ProgressStatus.NEUTRAL
            )

    def forget_bad_updates(self) -> None:
        """Reset unsuccessfull update counter and set the status to "neutral".

        Note that this method does NOT reset the best score.

        See also:
            `ProgressTracker.reset`
        """
        self._bad_counter = 0
        self._status = _ProgressStatus.NEUTRAL

    def reset(self) -> None:
        """Reset everything.

        See also:
            `ProgressTracker.forget_bad_updates`
        """
        self.forget_bad_updates()
        self._best_score = None


class Timer:
    """Measures time.

    Measures time elapsed since the first call to `~Timer.run` up to "now" minus pauses.

    Note:
        Measurements are performed via `time.perf_counter`.

    .. rubric:: Tutorial

    .. testcode::

        import time

        assert delu.Timer()() == 0.0

        timer = delu.Timer()
        timer.run()  # start
        time.sleep(0.01)
        assert timer()  # some time has passed

        timer.pause()
        elapsed = timer()
        time.sleep(0.01)
        assert timer() == elapsed  # time didn't change because the timer is on pause

        timer.run()  # resume
        time.sleep(0.01)
        assert timer() > elapsed

        timer.reset()
        assert timer() == 0.0

        with delu.Timer() as timer:
            time.sleep(0.01)
        # timer is on pause and timer() returns the time elapsed within the context

    `Timer` can be printed and formatted in a human-readable manner::

        timer = delu.Timer()
        timer.run()
        <let's assume that 3661 seconds have passed>
        print('Time elapsed:', timer)  # prints "Time elapsed: 1:01:01"
        assert str(timer) == f'{timer}' == '1:01:01'
        assert timer.format('%Hh %Mm %Ss') == '01h 01m 01s'

    `Timer` is pickle friendly:

    .. testcode::

        import pickle

        timer = delu.Timer()
        timer.run()
        time.sleep(0.01)
        timer.pause()
        old_value = timer()
        timer_bytes = pickle.dumps(timer)
        time.sleep(0.01)
        new_timer = pickle.loads(timer_bytes)
        assert new_timer() == old_value
    """

    # mypy cannot infer types from .reset(), so they must be given here
    _start_time: Optional[float]
    _pause_time: Optional[float]
    _shift: float

    def __init__(self) -> None:
        """Initialize self.

        Examples:
            .. testcode::

                timer = delu.Timer()
        """
        self.reset()

    def reset(self) -> None:
        """Reset the timer.

        Resets the timer to the initial state.
        """
        self._start_time = None
        self._pause_time = None
        self._shift = 0.0

    def run(self) -> None:
        """Start/resume the timer.

        If the timer is on pause, the method resumes the timer.
        If the timer is running, the method does nothing (i.e. it does NOT overwrite
        the previous pause time).
        """
        if self._start_time is None:
            self._start_time = time.perf_counter()
        elif self._pause_time is not None:
            self._shift -= time.perf_counter() - self._pause_time
            self._pause_time = None

    def pause(self) -> None:
        """Pause the timer.

        If the timer is running, the method pauses the timer.
        If the timer is already on pause, the method does nothing.

        Raises:
            AssertionError: if the timer is just created or just reset.
        """
        assert self._start_time is not None
        if self._pause_time is None:
            self._pause_time = time.perf_counter()

    def __call__(self) -> float:
        """Get the time elapsed since the start.

        Returns:
            Time elapsed.
        """
        if self._start_time is None:
            return self._shift
        now = self._pause_time or time.perf_counter()
        return now - self._start_time + self._shift

    def __str__(self) -> str:
        """Convert the timer to a string.

        Returns:
            The string representation of the timer's value rounded to seconds.
        """
        return str(datetime.timedelta(seconds=round(self())))

    def format(self, format_str: str) -> str:
        """Format the time elapsed since the start in a human-readable string.

        Args:
            format_str: the format string passed to `time.strftime`
        Returns:
            Filled ``format_str``.

        Example::

            timer = delu.Timer()
            <let's assume that 3661 seconds have passed>
            assert timer.format('%Hh %Mm %Ss') == '01h 01m 01s'
        """
        return time.strftime(format_str, time.gmtime(self()))

    def __enter__(self) -> 'Timer':
        """Measure time within a context.

        The method `Timer.run` is called regardless of the current state. On exit,
        `Timer.pause` is called.

        See also:
            `Timer.__exit__`

        Example:
            ..testcode::

                import time
                with delu.Timer() as timer:
                    time.sleep(0.01)
                elapsed = timer()
                assert elapsed > 0.01
                time.sleep(0.01)
                assert timer() == elapsed  # the timer is paused in __exit__
        """
        self.run()
        return self

    def __exit__(self, *args) -> bool:  # type: ignore
        """Leave the context and pause the timer.

        See `Timer.__enter__` for details and examples.

        See also:
            `Timer.__enter__`
        """
        self.pause()
        return False

    def __getstate__(self) -> Dict[str, Any]:
        return {'_shift': self(), '_start_time': None, '_pause_time': None}

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.__dict__.update(state)



"""Deep Learning Utilities for PyTorch users."""
__version__ = '0.0.14'

from . import cuda  # noqa: F401
from . import data  # noqa: F401
from . import hardware  # noqa: F401
from . import nn  # noqa: F401
from . import random  # noqa: F401
from ._monitoring import ProgressTracker, Timer  # noqa: F401
from ._stream import Stream  # noqa: F401
from ._tensor_ops import cat, concat, iter_batches, to  # noqa: F401
from ._utilities import evaluation, improve_reproducibility  # noqa: F401
from .data import collate  # noqa: F401
